package flow

import (
	"bytes"
	"encoding/json"
	"errors"
	"fmt"
	"reflect"
)

// Flow engine
type Flow struct {
	name   string
	reg    *Registry
	params map[string]interface{}
	// Will run on process
	op     Operation
	lastOp Operation
	Err    error

	runID int
}

// New Creates a new Flow app
func New(name string) *Flow {
	return &Flow{
		name:   name,
		reg:    nil, // should be internall
		params: map[string]interface{}{},
		op:     nil,
		lastOp: nil,
		Err:    nil,
		runID:  0,
	}
}

// SetRegistry sets the flow registry
func (f *Flow) SetRegistry(r *Registry) {
	if f.lastOp != nil {
		panic("Cannot set registry after operation added")
	}
	f.reg = r
}

//Op creates a new Operation
func (f *Flow) Op(name string, params ...interface{}) Operation {
	if f.Err != nil { // do nothing
		return nil
	}
	// Func to solve this aswell
	registry := f.reg
	if registry == nil {
		registry = globalRegistry
	}

	// Create
	op, err := registry.Get(name, f, params...)
	if err != nil {
		f.Err = fmt.Errorf("Creating: %s - %v", name, err)
	}
	f.lastOp = op
	return op
}

// SetParam sets a flow Param
func (f *Flow) SetParam(name string, value interface{}) {
	f.params[name] = value
}

// SetOutput this flow output
func (f *Flow) SetOutput(o Operation) {
	f.op = o
}

// Param return a Operation parameter
func (f *Flow) Param(name string) Operation {
	// Create a local param
	// and return operation ptr?
	return &opHandler{
		opSpec: func() OpSpec {
			return OpSpec{
				Name:       "param[" + name + "]",
				OutputType: reflect.TypeOf(f.params[name]),
			}
		},
		flow:     f,
		executor: func() Data { return f.params[name] },
		// Type?
	}
}

// Run several operations
// if operations is not specified
// we run the flow main output
func (f *Flow) Run(operations ...Operation) []Data {
	f.runID++

	if len(operations) == 0 {
		return []Data{f.Process()}
	}

	res := make([]Data, len(operations))
	for i, o := range operations {
		res[i] = o.Process() // Internal
	}
	return res
}

// Name Returns the flow name
func (f *Flow) Name() string {
	return f.name + "_flow"
}

// Process process the flow
func (f *Flow) Process() Data {
	// create Func to Solve operation
	op := f.op
	if op == nil {
		op = f.lastOp
	}
	if op == nil {
		f.Err = errors.New("Empty operation")
		return nil
	}

	return op.Process()
}

func (f *Flow) Spec() OpSpec {
	op := f.op
	if op == nil {
		op = f.lastOp
	}
	if op == nil {
		f.Err = errors.New("Empty operation")
		return OpSpec{}
	}

	return OpSpec{
		Name:       "flow",
		OutputType: op.Spec().OutputType,
	}
}

// try to serialize
func (f *Flow) String() string {
	ret := bytes.NewBuffer(nil)

	// Print flow parameters
	for k, v := range f.params {
		fmt.Fprintf(ret, "%s.Param[%s] - %v\n", f.name, k, v)
	}
	var printop func(o interface{}, prefix string)
	printop = func(o interface{}, prefix string) {
		if op, ok := o.(Operation); ok {
			spec := op.Spec()
			fmt.Fprintf(ret, "%s <- Name: %s (output: %v) (inputs: %v)\n", prefix, spec.Name, spec.OutputType, spec.InputType)
			switch op := o.(type) {
			case *opHandler:
				for _, in := range op.inputs {
					printop(in, prefix+"  ")
				}
			case *Flow:
				printop(op.op, prefix+"  ")
			}
			return
		}
		fmt.Fprintf(ret, "%s <- Constant: %v\n", prefix, o)
	}

	printop(f.op, "")

	return ret.String()
}

// Should appear something like
// {
//   'flowid': [
//			{
//			name:"vec",
//			inputs: [
//				{
//					name: "vec",
//					const: []float32{1,2,3},
//
//
//			]
//   ]
// }
func (f *Flow) ToJSON() string {
	s := &Serialize{[]Operation{}, map[int]opData{}}
	s.serialize(f.op)

	ret := bytes.NewBuffer(nil)
	e := json.NewEncoder(ret)
	e.SetIndent("", " ")
	e.Encode(s.dataOps)

	return ret.String()
}
