package flow

import (
	"errors"
	"fmt"
	"reflect"
)

type Type = reflect.Type
type Data = interface{}

// Maybe we can set input type
type OpSpec struct {
	Name       string
	InputType  []Type
	OutputType Type // For dynamics
}

// Operator
type Operation interface {
	Spec() OpSpec
	Init(params ...interface{}) error
	Process() Data // Or Value()
}

// Operation controller
type opHandler struct {
	flow   *Flow // use parent flow for certain things
	opSpec func() OpSpec
	inputs []Operation // should be an operation

	executor interface{} // Function that executes

	runID int
	cache interface{}

	// maybe cache
}

// Processes or fetches data
func (o *opHandler) Process() Data {
	if o.flow.runID == o.runID && o.cache != nil {
		return o.cache
	}

	callParam := make([]reflect.Value, len(o.inputs))
	for i, in := range o.inputs {
		switch ind := in.(type) {
		case Operation:
			callParam[i] = reflect.ValueOf(ind.Process())
		default:
			callParam[i] = reflect.ValueOf(ind)
		}
	}

	o.runID = o.flow.runID
	o.cache = reflect.ValueOf(o.executor).Call(callParam)[0].Interface()
	return o.cache
}
func (o *opHandler) Spec() OpSpec {
	return o.opSpec()
}

// Create a constructor for a func
func fromFunc(name string, fn interface{}) opFactory {
	fntyp := reflect.TypeOf(fn)

	// Constructor for func
	return func(flow *Flow, params ...Operation) (Operation, error) {
		// Do some checking first
		if fntyp.NumIn() != len(params) {
			return nil, fmt.Errorf("Invalid number of params expected: %d got %d", fntyp.NumIn(), len(params))
		}
		if fntyp.NumOut() == 0 {
			return nil, errors.New("Func it should have at least 1 output")
		}
		inputTypes := make([]Type, fntyp.NumIn())
		for i := 0; i < fntyp.NumIn(); i++ {
			inputTypes[i] = fntyp.In(i)
		}

		// Here
		op := &opHandler{
			opSpec: func() OpSpec {
				return OpSpec{
					InputType:  inputTypes,
					Name:       name,
					OutputType: fntyp.Out(0), // First output
				}
			},
			flow:     flow,
			inputs:   params,
			executor: fn,
		}
		return op, nil
	}
}
