package flow

// Data basic data interface
type Data = interface{}

// Operation data entry
type opEntry struct {
	name  string // Constructor name
	input []Ref  // Reference to input
}

// Flow structure
type Flow struct {
	registry   *Registry
	consts     []Data
	data       []Data
	operations []opEntry // Registered operations

	// Could be same with some indicator
}

// New create new flow
func New() *Flow {
	return &Flow{
		registry: globalRegistry,

		// Serializable
		operations: []opEntry{},
		data:       []Data{},
		consts:     []Data{},
	}
}

// SetRegistry sets the registry for this flow
func (f *Flow) SetRegistry(r *Registry) {
	f.registry = r
}

// Op add a new operation
func (f *Flow) Op(name string, params ...interface{}) Ref {
	refParams := make([]Ref, len(params))
	for i, p := range params {
		switch v := p.(type) {
		case Ref:
			refParams[i] = v
		default:
			refParams[i] = f.Const(v) // Should be a simple value and not a pointer
		}
	}
	f.operations = append(f.operations, opEntry{name, refParams})
	refID := len(f.operations) - 1
	return RefFunc{ref{f, refID}}
}

// Build the flow into a graph
func (f *Flow) Build() (*Graph, error) {
	g := &Graph{flow: f, operations: []operation{}}

	// Add operations to operator
	for _, op := range f.operations {
		fn, err := f.registry.Get(op.name)
		if err != nil {
			return nil, err
		}
		// If executor returns a function we can call it to instantiate
		// Create wrapper here
		newOp := &opFunc{flow: f, executor: fn}
		g.operations = append(g.operations, newOp)
		// Check what is the fn
		param := []interface{}{}
		for _, in := range op.input {
			switch ref := in.(type) {
			case RefFunc:
				param = append(param, g.operations[ref.id]) // operation already registered
			case RefConst:
				param = append(param, &opConst{f.consts[ref.id]})
			case RefVar:
				param = append(param, &opVar{flow: f, id: ref.id})
			}
		}

		err = newOp.Init(param...)
		if err != nil {
			return nil, err
		}
	}
	return g, nil
}

// Const new constant
func (f *Flow) Const(v interface{}) Ref {
	// Add a ref
	f.consts = append(f.consts, v)
	refID := len(f.consts) - 1
	return RefConst{ref{f, refID}}
}

// Variable new variable with initial value
func (f *Flow) Variable(initial interface{}) RefVar {
	// Add a ref
	f.data = append(f.data, initial)
	refID := len(f.data) - 1
	return RefVar{ref{f, refID}}
}
